//
// TeensyLC programmable joystick mouse v0.1 BETA
//

// Requires Teensyduino to be installed on top of Arduino
#include <stdint.h>
#include <Bounce.h>
#include <Encoder.h>
#include <EEPROM.h>

// ========================================================================================
// PIN ASSIGNMENTS
// ========================================================================================

// joystick module pins

#define PIN_BUTTON_JOY 16
#define PIN_JOY_X      A1  // 15
#define PIN_JOY_Y      A0  // 14

// rotary encoder pins (scrollwheel)

#define PIN_ENCODER_L  20
#define PIN_ENCODER_R  21

// button pins

#define PIN_BUTTON_TOP    9
#define PIN_BUTTON_WHEEL  10
#define PIN_BUTTON_BOTTOM 11
#define PIN_BUTTON_MODE   12

// led pins

#define PIN_LED_1  0
#define PIN_LED_2  1
#define PIN_LED_3  2

// buzzer pins

#define PIN_BUZZER  3

// vibrator pins

#define PIN_RUMBLE  4

// ========================================================================================
// DEFINITIONS
// ========================================================================================

// device operation modes

#define MODE_BOOTUP           0    // startup
#define MODE_NORMAL           1    // emulate mouse
#define MODE_CALIBRATE_CENTRE 2    // joystick calibration (set centre)
#define MODE_CALIBRATE_BOUNDS 3    // joystick calibration (set min/max)
#define MODE_PROGRAM          4    // write presets
#define MODE_READ_SETTINGS    5    // read presets

// event handler value defintions

#define HANDLER_NOTHING              0b00000000
#define HANDLER_CURSOR_POS_X         0b00000001
#define HANDLER_CURSOR_NEG_X         0b00000010
#define HANDLER_CURSOR_POS_Y         0b00000011
#define HANDLER_CURSOR_NEG_Y         0b00000100
#define HANDLER_SCROLL_VERT_POS      0b00000101
#define HANDLER_SCROLL_VERT_NEG      0b00000110
#define HANDLER_SCROLL_HORIZ_POS     0b00000111
#define HANDLER_SCROLL_HORIZ_NEG     0b00001000
#define HANDLER_MOUSE_CLICK_1        0b00001001
#define HANDLER_MOUSE_CLICK_2        0b00001010
#define HANDLER_MOUSE_CLICK_3        0b00001011
#define HANDLER_MOUSE_DBL_CLICK_1    0b00001100
#define HANDLER_MOUSE_DBL_CLICK_2    0b00001101
#define HANDLER_MOUSE_DBL_CLICK_3    0b00001110
#define HANDLER_MOUSE_TOGGLE_1       0b00001111
#define HANDLER_MOUSE_TOGGLE_2       0b00010000
#define HANDLER_MOUSE_TOGGLE_3       0b00010001
#define HANDLER_MOUSE_CLEAR_1        0b00010010
#define HANDLER_MOUSE_CLEAR_2        0b00010011
#define HANDLER_MOUSE_CLEAR_3        0b00010100
#define HANDLER_MOUSE_SET_1          0b00010101
#define HANDLER_MOUSE_SET_2          0b00010110
#define HANDLER_MOUSE_SET_3          0b00010111
#define HANDLER_SCROLL_SWAP          0b00011000
#define HANDLER_CLICK_SWAP           0b00011001
#define HANDLER_MOUSE_CLEAR_ALL      0b00011010
#define HANDLER_NEXT_PRESET          0b00011111
#define HANDLER_MASK                 0b00011111

#define HANDLER_EXTRA_NONE           0b00000000
#define HANDLER_EXTRA_RUMBLE         0b00100000
#define HANDLER_EXTRA_BEEP           0b01000000
#define HANDLER_EXTRA_BLINK          0b10000000
#define HANDLER_EXTRA_MASK           0b11100000

// bit flags

#define FLAG_NONE            0
#define FLAG_LED_ENABLED     1
#define FLAG_BEEP_ENABLED    2
#define FLAG_RUMBLE_ENABLED  4
#define FLAG_PRESET_ENABLED  128

// axis interpolation settings

#define AXIS_X_INT_JUMP    0x01
#define AXIS_X_INT_LINEAR  0x02
#define AXIS_X_INT_SMOOTH  0x03
#define AXIS_Y_INT_JUMP    0x10
#define AXIS_Y_INT_LINEAR  0x20
#define AXIS_Y_INT_SMOOTH  0x30
#define AXIS_X_INT_MASK    0x0F
#define AXIS_Y_INT_MASK    0xF0

// memory structures

struct preset_t
{
  // event handler assignments
  uint8_t axis_x_pos;
  uint8_t axis_x_neg;
  uint8_t axis_y_pos;
  uint8_t axis_y_neg;
  uint8_t axis_recentre;
  uint8_t axis_button;
  uint8_t top_button;
  uint8_t scroll_button;
  uint8_t bottom_button;
  uint8_t mode_button;
  uint8_t scroll_neg;
  uint8_t scroll_pos;

  // settings
  uint8_t bit_flags;                 // enabled, beep_on, rumble_on, led_on
  uint8_t axis_interpolation;        // square, linear, curve /axis
  uint8_t x_axis_speed[8];
  uint8_t y_axis_speed[8];
  uint8_t scroll_vertical_rate;
  uint8_t scroll_horizontal_rate;
  uint8_t scroll_delay;
};
// 33 bytes

struct settings_t
{
  uint8_t current_preset;
  uint8_t bit_flags;

  int16_t joy_deadzone_x;
  int16_t joy_deadzone_y;
  int16_t joy_x_min;
  int16_t joy_x_max;
  int16_t joy_y_min;
  int16_t joy_y_max;
  int16_t joy_x_centre;
  int16_t joy_y_centre;
  
  int16_t beep_tone;
  uint8_t beep_length;

  uint8_t rumble_power;
  uint8_t rumble_length;
};
// 24 bytes

struct setup_t 
{
  settings_t settings;
  preset_t presets[3];
};

union setup_memory_t 
{
  setup_t setup;
  byte memory[128];
};

// ========================================================================================
// GLOBALS
// ========================================================================================
setup_memory_t s;

uint8_t mode = MODE_BOOTUP;
int32_t wheel_position = 0;
int32_t new_wheel_position;
int32_t joy_x;
int32_t joy_y;
int32_t joy_delta_x;
int32_t joy_delta_y;
uint8_t joy_state;                   // track joystick usage
uint8_t button_state[3] = {0,0,0};   // track emulated button states
uint8_t mouse_click_swap = 0;        // swapped single-click/double-click events
uint8_t scroll_orientation_swap = 0; // swapped horizontal/vertical scroll
uint16_t blink_countdown = 0;        // count to blank leds
uint16_t rumble_countdown = 0;       // count to rumble
uint8_t calibrate_blink = 0;           // calibration blink notification mode

Encoder wheel(PIN_ENCODER_L, PIN_ENCODER_R);
Bounce joy_button = Bounce(PIN_BUTTON_JOY, 10);
Bounce top_button = Bounce(PIN_BUTTON_TOP, 10);
Bounce wheel_button = Bounce(PIN_BUTTON_WHEEL, 10);
Bounce bottom_button = Bounce(PIN_BUTTON_BOTTOM, 10);
Bounce mode_button = Bounce(PIN_BUTTON_MODE, 10);

// ========================================================================================
// SETUP
// ========================================================================================
void setup() 
{
  // configure pins
  pinMode(PIN_BUTTON_JOY, INPUT_PULLUP);
  pinMode(PIN_JOY_X, INPUT);
  pinMode(PIN_JOY_Y, INPUT);

  pinMode(PIN_BUTTON_TOP, INPUT_PULLUP);
  pinMode(PIN_BUTTON_WHEEL, INPUT_PULLUP);
  pinMode(PIN_BUTTON_BOTTOM, INPUT_PULLUP);
  pinMode(PIN_BUTTON_MODE, INPUT_PULLUP);

  pinMode(PIN_LED_1, OUTPUT);
  pinMode(PIN_LED_2, OUTPUT);
  pinMode(PIN_LED_3, OUTPUT);

  // start serial
  Serial.begin(9600);
}

// ========================================================================================
// MAIN LOOP
// ========================================================================================
void loop() 
{
  if (mode == MODE_BOOTUP)
  {
    delay(1000);
    read_settings();
    mode = MODE_NORMAL;
  }

  // ------------------------------------------------------------------------
  if (mode == MODE_PROGRAM)
  {
    // write settings memory from serial port
    Serial.readBytes((char*)s.memory, 128);
    Serial.flush();
    
    // (simple!) checksum
    uint16_t sum = 0;
    for (uint8_t a = 0; a != 126; ++a)
    {
      sum += s.memory[a];
    }
    if (sum == (s.memory[126] | (s.memory[127] << 8)))
    {
      // save to EEPROM
      save_settings();
      Serial.write('K');
    }
    else
    {
      // rollback to last saved EEPROM
      read_settings();
      Serial.write('E');
    }
    // return to normal mode
    mode = MODE_NORMAL;
  }

  if (mode == MODE_READ_SETTINGS)
  {
    // dump settings memory to serial port
    Serial.write(s.memory,128);
    Serial.flush();
    mode = MODE_NORMAL;
  }

  // ------------------------------------------------------------------------
  if (mode == MODE_CALIBRATE_CENTRE)
  {
    // blink leds to indicate calibration mode
    if (--blink_countdown == 0)
    {
      blink_countdown = 100;
      calibrate_blink = !calibrate_blink;
      update_leds_calibrate();
    }
    
    // read joystick position
    joy_x = analogRead(PIN_JOY_X);
    joy_y = analogRead(PIN_JOY_Y);
    
    // check mode button
    mode_button.update();
    
    // finish centre calibration
    if (mode_button.risingEdge())
    {
      // set centre location
      s.setup.settings.joy_x_centre = joy_x;
      s.setup.settings.joy_y_centre = joy_y;

      // reset bounds
      s.setup.settings.joy_x_min = s.setup.settings.joy_x_centre;
      s.setup.settings.joy_x_max = s.setup.settings.joy_x_centre;
      s.setup.settings.joy_y_min = s.setup.settings.joy_y_centre;
      s.setup.settings.joy_y_max = s.setup.settings.joy_y_centre;
      
      Serial.write('B');
      mode = MODE_CALIBRATE_BOUNDS;
    }
  }

  // ------------------------------------------------------------------------
  if (mode == MODE_CALIBRATE_BOUNDS)
  {
    // blink leds to indicate calibration mode
    if (--blink_countdown == 0)
    {
      blink_countdown = 100;
      calibrate_blink = !calibrate_blink;
      update_leds_calibrate();
    }
    
    // read joystick position
    joy_x = analogRead(PIN_JOY_X);
    joy_y = analogRead(PIN_JOY_Y);

    // set bounds to min/max readings
    if (s.setup.settings.joy_x_min > joy_x)
    {
      s.setup.settings.joy_x_min = joy_x;
    }
    if (s.setup.settings.joy_x_max < joy_x)
    {
      s.setup.settings.joy_x_max = joy_x;
    }
    
    if (s.setup.settings.joy_y_min > joy_y)
    {
      s.setup.settings.joy_y_min = joy_y;
    }
    if (s.setup.settings.joy_y_max < joy_y)
    {
      s.setup.settings.joy_y_max = joy_y;
    }

    // check mode button
    mode_button.update();
    
    // finish calibration
    if (mode_button.risingEdge())
    {
      // reset leds
      blink_countdown = 0;
      calibrate_blink = 0;
      update_leds_calibrate();

      // save to EEPROM
      save_settings();
      Serial.write('N');
      mode = MODE_NORMAL;
    }
  }

  // ------------------------------------------------------------------------
  if (mode == MODE_NORMAL)
  {
    // check serial port for commands, change modes
    if (Serial.available() > 0)
    {
      char cmd = Serial.read();
      if (cmd == 'C')
      {
        Serial.write('A');
        calibrate_blink = 1;
        blink_countdown = 100;
        mode = MODE_CALIBRATE_CENTRE;
      }
      if (cmd == 'R')
      {
        mode = MODE_READ_SETTINGS;
      }
      if (cmd == 'T')
      {
        // test connection, repsond with acknowledgement
        Serial.write('A');
      }
      if (cmd == 'P')
      {
        mode = MODE_PROGRAM;
      }
    }
    
    // read joystick position
    joy_x = analogRead(PIN_JOY_X);
    joy_y = analogRead(PIN_JOY_Y);
    joy_delta_x = (joy_x-s.setup.settings.joy_x_centre);
    joy_delta_y = (joy_y-s.setup.settings.joy_y_centre);
  
    // joystick events
    if (joy_delta_x > s.setup.settings.joy_deadzone_x)
    {
      //Mouse.move(2,0);
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_x_pos, 0);
      joy_state = 1;
    }
    if (joy_delta_x < -s.setup.settings.joy_deadzone_x)
    {
      //Mouse.move(-2,0);
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_x_neg, 0);
      joy_state = 1;
    }
    if (joy_delta_y > s.setup.settings.joy_deadzone_y)
    {
      //Mouse.move(0,-2);
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_y_neg, 0);
      joy_state = 1;
    }
    if (joy_delta_y < -s.setup.settings.joy_deadzone_y)
    {
      //Mouse.move(0,2);
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_y_pos, 0);
      joy_state = 1;
    }
    
    // joystick recentred after being moved out of deadzone?
    if (joy_state == 1 && 
      joy_delta_x < s.setup.settings.joy_deadzone_x && joy_delta_x > -s.setup.settings.joy_deadzone_x &&
      joy_delta_y < s.setup.settings.joy_deadzone_y && joy_delta_y > -s.setup.settings.joy_deadzone_y)
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_recentre, 0);
      joy_state = 0;
    }
  
    // get wheel position
    new_wheel_position = wheel.read();
    
    // wheel events
    if (new_wheel_position > wheel_position) 
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].scroll_pos, 0);
      wheel_position = new_wheel_position;
    }
    else if (new_wheel_position < wheel_position) 
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].scroll_neg, 0);
      wheel_position = new_wheel_position;
    }
  
    // refresh button status
    joy_button.update();
    top_button.update();
    wheel_button.update();
    bottom_button.update();
    mode_button.update();
  
    // button events
    if (top_button.fallingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].top_button, 1);
    }
    if (top_button.risingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].top_button, 0);
    }
    if (wheel_button.fallingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].scroll_button, 1);
    }
    if (wheel_button.risingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].scroll_button, 0);
    }  
    if (bottom_button.fallingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].bottom_button, 1);
    }
    if (bottom_button.risingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].bottom_button, 0);
    }
    if (joy_button.fallingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_button, 1);
    }
    if (joy_button.risingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].axis_button, 0);
    }
    if (mode_button.fallingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].mode_button, 1);
    }
    if (mode_button.risingEdge())
    {
      handle_event(s.setup.presets[s.setup.settings.current_preset].mode_button, 0);
    }
  
    // refresh
    update_leds();
    update_rumble();
  
  } // normal mode
  
  delay(10);
}

// ========================================================================================
// EVENT HANDLERS
// ========================================================================================
void handle_event(uint8_t handler_code, uint8_t fe)
{
  switch (handler_code & HANDLER_MASK)
  {
    case HANDLER_NOTHING:
      return;
      break;
    case HANDLER_CURSOR_POS_X:
      Mouse.move(get_interpolation_x(),0);
      break;
    case HANDLER_CURSOR_NEG_X:
      Mouse.move(-get_interpolation_x(),0);
      break;
    case HANDLER_CURSOR_POS_Y:
      Mouse.move(0,get_interpolation_y());
      break;
    case HANDLER_CURSOR_NEG_Y:
      Mouse.move(0,-get_interpolation_y());
      break;
    case HANDLER_SCROLL_VERT_POS:
      if (!scroll_orientation_swap)
      {
        Mouse.scroll(s.setup.presets[s.setup.settings.current_preset].scroll_vertical_rate);
      }
      else
      {
        Mouse.scroll(0,s.setup.presets[s.setup.settings.current_preset].scroll_horizontal_rate);
      }
      delay(s.setup.presets[s.setup.settings.current_preset].scroll_delay);
      break;
    case HANDLER_SCROLL_VERT_NEG:
      if (!scroll_orientation_swap)
      {
        Mouse.scroll(-s.setup.presets[s.setup.settings.current_preset].scroll_vertical_rate);
      }
      else
      {
        Mouse.scroll(0,-s.setup.presets[s.setup.settings.current_preset].scroll_horizontal_rate);
      }
      delay(s.setup.presets[s.setup.settings.current_preset].scroll_delay);
      break;
    case HANDLER_SCROLL_HORIZ_POS:
      if (!scroll_orientation_swap)
      {
        Mouse.scroll(0,s.setup.presets[s.setup.settings.current_preset].scroll_horizontal_rate);
      }
      else
      {
        Mouse.scroll(s.setup.presets[s.setup.settings.current_preset].scroll_vertical_rate);
      }
      delay(s.setup.presets[s.setup.settings.current_preset].scroll_delay);
      break;
    case HANDLER_SCROLL_HORIZ_NEG:
      if (!scroll_orientation_swap)
      {
        Mouse.scroll(0,-s.setup.presets[s.setup.settings.current_preset].scroll_horizontal_rate);
      }
      else
      {
        Mouse.scroll(-s.setup.presets[s.setup.settings.current_preset].scroll_vertical_rate);
      }
      delay(s.setup.presets[s.setup.settings.current_preset].scroll_delay);
      break;
    case HANDLER_MOUSE_CLICK_1:
      if (!mouse_click_swap)
      {
        single_click(MOUSE_LEFT, fe, 0);
      }
      else
      {
        double_click(MOUSE_LEFT, fe, 0);
      }
      break;
    case HANDLER_MOUSE_CLICK_2:
      if (!mouse_click_swap)
      {
        single_click(MOUSE_RIGHT, fe, 1);
      }
      else
      {
        double_click(MOUSE_RIGHT, fe, 1);
      }
      break;
    case HANDLER_MOUSE_CLICK_3:
      if (!mouse_click_swap)
      {
        single_click(MOUSE_MIDDLE, fe, 2);
      }
      else
      {
        double_click(MOUSE_MIDDLE, fe, 2);
      }
      break;
    case HANDLER_MOUSE_DBL_CLICK_1:
      if (mouse_click_swap)
      {
        single_click(MOUSE_LEFT, fe, 0);
      }
      else
      {
        double_click(MOUSE_LEFT, fe, 0);
      }
      break;    
    case HANDLER_MOUSE_DBL_CLICK_2:
      if (mouse_click_swap)
      {
        single_click(MOUSE_RIGHT, fe, 1);
      }
      else
      {
        double_click(MOUSE_RIGHT, fe, 1);
      }
      break;
    case HANDLER_MOUSE_DBL_CLICK_3:
      if (mouse_click_swap)
      {
        single_click(MOUSE_MIDDLE, fe, 2);
      }
      else
      {
        double_click(MOUSE_MIDDLE, fe, 2);
      }
      break;
    case HANDLER_MOUSE_TOGGLE_1:
      if (fe)
      {
        if (button_state[0])
        {
          Mouse.release(MOUSE_LEFT);
          button_state[0] = 0;
        }
        else
        {
          Mouse.press(MOUSE_LEFT);
          button_state[0] = 1;
        }
      }
      break;
    case HANDLER_MOUSE_TOGGLE_2:
      if (fe)
      {
        if (button_state[1])
        {
          Mouse.release(MOUSE_RIGHT);
          button_state[1] = 0;
        }
        else
        {
          Mouse.press(MOUSE_RIGHT);
          button_state[1] = 1;
        }
      }
      break;
    case HANDLER_MOUSE_TOGGLE_3:
      if (fe)
      {
        if (button_state[2])
        {
          Mouse.release(MOUSE_MIDDLE);
          button_state[2] = 0;
        }
        else
        {
          Mouse.press(MOUSE_MIDDLE);
          button_state[2] = 1;
        }
      }
      break;
    case HANDLER_MOUSE_CLEAR_1:
      if (!fe)
      {
        Mouse.release(MOUSE_LEFT);
        button_state[0] = 0;
      }
      break;
    case HANDLER_MOUSE_CLEAR_2:
      if (!fe)
      {
        Mouse.release(MOUSE_RIGHT);
        button_state[1] = 0;
      }
      break;
    case HANDLER_MOUSE_CLEAR_3:
      if (!fe)
      {
        Mouse.release(MOUSE_MIDDLE);
        button_state[2] = 0;
      }
      break;
    case HANDLER_MOUSE_SET_1:
      if (!fe)
      {
        Mouse.press(MOUSE_LEFT);
        button_state[0] = 1;
      }
      break;
    case HANDLER_MOUSE_SET_2:
      if (!fe)
      {
        Mouse.press(MOUSE_RIGHT);
        button_state[1] = 1;
      }    
      break;
    case HANDLER_MOUSE_SET_3:
      if (!fe)
      {
        Mouse.press(MOUSE_MIDDLE);
        button_state[2] = 1;
      }    
      break;
    case HANDLER_SCROLL_SWAP:
      if (fe)
      {
        if (scroll_orientation_swap)
        {
          scroll_orientation_swap = 0;
        }
        else
        {
          scroll_orientation_swap = 1;
        }
      }
      break;
    case HANDLER_CLICK_SWAP:
      if (fe)
      {
        if (mouse_click_swap)
        {
          mouse_click_swap = 0;
        }
        else
        {
          mouse_click_swap = 1;
        }
      }
      break;
    case HANDLER_MOUSE_CLEAR_ALL:
      if (!fe)
      {
        Mouse.release(MOUSE_LEFT);
        Mouse.release(MOUSE_RIGHT);
        Mouse.release(MOUSE_MIDDLE);
        button_state[0] = 0;
        button_state[1] = 0;
        button_state[2] = 0;
      }
      break;
    case HANDLER_NEXT_PRESET:
      if (fe)
      {
        // reset states
        Mouse.release(MOUSE_LEFT);
        Mouse.release(MOUSE_RIGHT);
        Mouse.release(MOUSE_MIDDLE);
        button_state[0] = 0;
        button_state[1] = 0;
        button_state[2] = 0;
        mouse_click_swap = 0;
        scroll_orientation_swap = 0;

        // move to next enabled preset, give up if we loop
        uint32_t count = 0;
        do
        {
          if (++s.setup.settings.current_preset == 3)
          {
            s.setup.settings.current_preset = 0;
          }
          if (++count==3)
          {
            break;
          }
        } while ((s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_PRESET_ENABLED) != FLAG_PRESET_ENABLED);
      }
      break;
  }

  // additional effects on handler?
  if (!fe)
  {
    if ((handler_code & HANDLER_EXTRA_RUMBLE) == HANDLER_EXTRA_RUMBLE)
    {
      // rumble, check if global enable, preset enable
      if ((s.setup.settings.bit_flags & FLAG_RUMBLE_ENABLED) == FLAG_RUMBLE_ENABLED && (s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_RUMBLE_ENABLED) == FLAG_RUMBLE_ENABLED)
      {
        analogWrite(PIN_RUMBLE, s.setup.settings.rumble_power);
        rumble_countdown = s.setup.settings.rumble_length;
      }
    }
    if ((handler_code & HANDLER_EXTRA_BEEP) == HANDLER_EXTRA_BEEP)
    {
      // beep, check if global enable, preset enable
      if ((s.setup.settings.bit_flags & FLAG_BEEP_ENABLED) == FLAG_BEEP_ENABLED && (s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_BEEP_ENABLED) == FLAG_BEEP_ENABLED)
      {
        tone(PIN_BUZZER, s.setup.settings.beep_tone, s.setup.settings.beep_length);
      }
    }
    if ((handler_code & HANDLER_EXTRA_BLINK) == HANDLER_EXTRA_BLINK)
    {
      // blink, check if global enable, preset enable
      if ((s.setup.settings.bit_flags & FLAG_LED_ENABLED) == FLAG_LED_ENABLED && (s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_LED_ENABLED) == FLAG_LED_ENABLED)
      {
        blink_countdown = 10;
      }
    }
  }  
}

void single_click(uint8_t mouse_button, uint8_t fe, uint8_t index)
{
  if (fe)
  {
    Mouse.press(mouse_button);
    button_state[index] = 1;
  }
  else
  {
    Mouse.release(mouse_button);
    button_state[index] = 0;
  }
}

void double_click(uint8_t mouse_button, uint8_t fe, uint8_t index)
{
  if (fe)
  {
    button_state[index] = 1;
    Mouse.press(mouse_button);
    delay(100);
    Mouse.release(mouse_button);
    delay(100);
    Mouse.press(mouse_button);
    delay(100);
    Mouse.release(mouse_button);
    button_state[index] = 0;
  }
  else
  {
    button_state[index] = 0;
  }
}

// ========================================================================================
// INTERPOLATION
// ========================================================================================
int32_t get_interpolation_x()
{
  int32_t range = 0;
  int32_t delta = joy_delta_x;
  if (delta > 0)
  {
    range = s.setup.settings.joy_x_max-s.setup.settings.joy_x_centre;
  }
  else
  {
    range = s.setup.settings.joy_x_centre-s.setup.settings.joy_x_min;
    delta *= -1;
  }

  if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_X_INT_MASK) == AXIS_X_INT_JUMP)
  {
    return jump_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].x_axis_speed);
  }
  else if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_X_INT_MASK) == AXIS_X_INT_LINEAR)
  {
    return linear_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].x_axis_speed);
  }
  else if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_X_INT_MASK) == AXIS_X_INT_SMOOTH)
  {
    return smooth_step_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].x_axis_speed);
  }

  return 0;
}

int32_t get_interpolation_y()
{
  int32_t range = 0;
  int32_t delta = joy_delta_y;
  if (delta > 0)
  {
    range = s.setup.settings.joy_y_max-s.setup.settings.joy_y_centre;
  }
  else
  {
    range = s.setup.settings.joy_y_centre-s.setup.settings.joy_y_min;
    delta *= -1;
  }

  if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_Y_INT_MASK) == AXIS_Y_INT_JUMP)
  {
    return jump_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].y_axis_speed);
  }
  else if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_Y_INT_MASK) == AXIS_Y_INT_LINEAR)
  {
    return linear_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].y_axis_speed);
  }
  else if ((s.setup.presets[s.setup.settings.current_preset].axis_interpolation & AXIS_Y_INT_MASK) == AXIS_Y_INT_SMOOTH)
  {
    return smooth_step_interpolation(delta, range, s.setup.presets[s.setup.settings.current_preset].y_axis_speed);
  }

  return 0;
}

int32_t jump_interpolation(int32_t delta, int32_t range, uint8_t *axis_speed)
{
  int32_t index = (int)(((delta*1.0f)/range)*7);
  return axis_speed[index];
}

int32_t linear_interpolation(int32_t delta, int32_t range, uint8_t *axis_speed)
{
  float float_index = ((delta*1.0f)/range)*7;
  int32_t index = (int32_t)(float_index);
  int32_t first_value = axis_speed[index];
  int32_t second_value = axis_speed[index+1];
  
  int32_t diff = second_value - first_value;
  float percent = float_index-index;
  return (int32_t)(first_value+(diff*percent));
}

int32_t smooth_step_interpolation(int32_t delta, int32_t range, uint8_t *axis_speed)
{
  float float_index = ((delta*1.0f)/range)*7;
  int32_t index = (int32_t)(float_index);
  int32_t first_value = axis_speed[index];
  int32_t second_value = axis_speed[index+1];
  int32_t diff = second_value - first_value;
  float percent = float_index-index;
  float ss_percent = smooth_step(0.0f, 1.0f, percent);
  return (int32_t)(first_value+(diff*ss_percent));  
}

float smooth_step(float edge0, float edge1, float x)
{
  x = clamp((x-edge0)/(edge1-edge0), 0.0f, 1.0f);
  return x * x * (3 - 2 * x);
}

float clamp(float x, float lower_limit, float upper_limit)
{
  if (x < lower_limit)
      x = lower_limit;
  if (x > upper_limit)
      x = upper_limit;
  return x;  
}

// ========================================================================================
// VISUAL/HAPTICS 
// ========================================================================================
void update_leds()
{
  if ((s.setup.settings.bit_flags & FLAG_LED_ENABLED) == FLAG_LED_ENABLED && (s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_LED_ENABLED) == FLAG_LED_ENABLED)
  {
    if (blink_countdown > 0)
    {
      --blink_countdown;
      digitalWrite(PIN_LED_1, 0);
      digitalWrite(PIN_LED_2, 0);
      digitalWrite(PIN_LED_3, 0);
    }
    else
    {
       digitalWrite(PIN_LED_1, (s.setup.settings.current_preset == 0 ? 1 : 0));
       digitalWrite(PIN_LED_2, (s.setup.settings.current_preset == 1 ? 1 : 0));
       digitalWrite(PIN_LED_3, (s.setup.settings.current_preset == 2 ? 1 : 0));
    }
  }
  else
  {
    digitalWrite(PIN_LED_1, 0);
    digitalWrite(PIN_LED_2, 0);
    digitalWrite(PIN_LED_3, 0);
  }
}

void update_rumble()
{
  if ((s.setup.settings.bit_flags & FLAG_RUMBLE_ENABLED) == FLAG_RUMBLE_ENABLED && (s.setup.presets[s.setup.settings.current_preset].bit_flags & FLAG_RUMBLE_ENABLED) == FLAG_RUMBLE_ENABLED)
  {
    if (rumble_countdown > 0)
    {
      --rumble_countdown;
    }
    else
    {
      analogWrite(PIN_RUMBLE, 0);
    }
  }
}

void update_leds_calibrate()
{
  if (calibrate_blink == 0)
  {
    digitalWrite(PIN_LED_1, 0);
    digitalWrite(PIN_LED_2, 0);
    digitalWrite(PIN_LED_3, 0);
  }
  else
  {
    digitalWrite(PIN_LED_1, 1);
    digitalWrite(PIN_LED_2, 1);
    digitalWrite(PIN_LED_3, 1);
  }
}

// ========================================================================================
// EEPROM
// ========================================================================================
void read_settings()
{
  for (uint8_t a = 0; a != 128; ++a)
  {
    s.memory[a] = EEPROM.read(a);
  }
}

void save_settings()
{
  for (uint8_t a = 0; a != 128; ++a)
  {
    // only write to eeprom if the value has changed to minimise wear
    byte old_value = EEPROM.read(a);
    if (old_value != s.memory[a])
    {
      EEPROM.write(a,s.memory[a]);
    }
  }
}
